# -*- coding: utf-8 -*-
from __future__ import absolute_import
import re
import math

from pyzxcvbn import scoring
from .adjacency_graphs import adjacency_graphs
from .frequency_lists import frequency_lists
from six.moves import filter
from six.moves import range


def build_ranked_dict(ordered_list):
    """Return ranked dict of word list
    :param list ordered_list:
    :rtype: dict
    """

    result = {}
    i = 1
    for word in ordered_list:
        result[word] = i
        i += 1
    return result

RANKED_DICTIONARIES = {
    "passwords": build_ranked_dict(frequency_lists["passwords"]),
    "english": build_ranked_dict(frequency_lists["english"]),
    "surnames": build_ranked_dict(frequency_lists["surnames"]),
    "male_names": build_ranked_dict(frequency_lists["male_names"]),
    "female_names": build_ranked_dict(frequency_lists["female_names"])
}

GRAPHS = {
    "qwerty": adjacency_graphs["qwerty"],
    "dvorak": adjacency_graphs["dvorak"],
    "keypad": adjacency_graphs["keypad"],
    "mac_keypad": adjacency_graphs["mac_keypad"]
}

SEQUENCES = {
    "lower": "abcdefghijklmnopqrstuvwxyz",
    "upper": "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
    "digits": "0123456789"
}

L33T_TABLE = {
    "a": ['4', '@'],
    "b": ['8'],
    "c": ['(', '{', '[', '<'],
    "e": ['3'],
    "g": ['6', '9'],
    "i": ['1', '!', '|'],
    "l": ['1', '|', '7'],
    "o": ['0'],
    "s": ['$', '5'],
    "t": ['+', '7'],
    "x": ['%'],
    "z": ['2']
}

REGEXEN = {
    "alphanumeric": "[a-zA-Z0-9]{2,}",
    "alpha":        "[a-zA-Z]{2,}",
    "alpha_lower":  "[a-z]{2,}",
    "alpha_upper":  "[A-Z]{2,}",
    "digits":       "\d{2,}",
    "symbols":      "[\W_]{2,}",  # includes non-latin unicode chars
    "recent_year":  "19\d\d|200\d|201\d",
}

REGEX_PRECEDENCE = {
    "alphanumeric": 0,
    "alpha":        1,
    "alpha_lower":  2,
    "alpha_upper":  2,
    "digits":       2,
    "symbols":      2,
    "recent_year":  3
}

DATE_MAX_YEAR = 2050
DATE_MIN_YEAR = 1000
DATE_SPLITS = {
    4: [      # for length-4 strings, eg 1191 or 9111, two ways to split:
        [1, 2],  # 1 1 91 (2nd split starts at index 1, 3rd at index 2)
        [2, 3]   # 91 1 1
    ],
    5: [
        [1, 3],  # 1 11 91
        [2, 3]   # 11 1 91
    ],
    6: [
        [1, 2],  # 1 1 1991
        [2, 4],  # 11 11 91
        [4, 5]   # 1991 1 1
    ],
    7: [
        [1, 3],  # 1 11 1991
        [2, 3],  # 11 1 1991
        [4, 5],  # 1991 1 11
        [4, 6]   # 1991 11 1
    ],
    8: [
        [2, 4],  # 11 11 1991
        [4, 6]   # 1991 11 11
    ]
}


def is_empty(obj):
    return len([k for k in obj]) == 0


def translate(string, c_map):
    """Translate according to character mapping table
    :param str string:
    :param dict c_map:
    :rtype: str
    """
    mapped_string = ""
    for c in string:
        if c in c_map:
            mapped_string += c_map[c]
        else:
            mapped_string += c
    return mapped_string


def mod(n, m):
    return ((n % m) + m) % m


def omnimatch(password):
    """Apply all match functions
    :param str password:
    :rtype: list
    """
    matches_all = []
    matchers = [
        dictionary_match,
        reverse_dictionary_match,
        l33t_match,
        spatial_match,
        repeat_match,
        sequence_match,
        regex_match,
        date_match
    ]

    for matcher in matchers:
        matches = matcher(password)
        matches_all += matches
    return sorted(matches_all, key=lambda x: (x['i'], x['j']))


def dictionary_match(password, _ranked_dictionaries=RANKED_DICTIONARIES):
    """

    :param str password:
    :param dict _ranked_dictionaries:
    :return:
    """
    matches = []
    length = len(password)

    password_lower = password.lower()

    for dictionary_name, ranked_dict in _ranked_dictionaries.items():
        for i in range(length):
            for j in range(length):
                token = password_lower[i:j+1]
                if token in ranked_dict:
                    word = token
                    rank = ranked_dict[word]
                    matches.append({
                        "pattern": "dictionary",
                        "i": i,
                        "j": j,
                        "token": password[i:j+1],
                        "matched_word": word,
                        "rank": rank,
                        "dictionary_name": dictionary_name,
                        "reversed": False
                    })

    return sorted(matches, key=lambda x: (x['i'], x['j']))


def reverse_dictionary_match(password, _ranked_dictionaries=RANKED_DICTIONARIES):
    reversed_password = password[::-1]
    matches = dictionary_match(reversed_password, _ranked_dictionaries)
    for match in matches:
        match["token"] = match["token"][::-1]
        match["reversed"] = True
        match["i"], match["j"] = (len(password) - 1 - match["j"], len(password) - 1 - match["i"])

    return sorted(matches, key=lambda x: (x['i'], x['j']))


def set_user_input_dictionary(ordered_list):
    """Set user-defined dictionary
    :param list ordered_list:
    :return: None
    """
    RANKED_DICTIONARIES["user_inputs"] = build_ranked_dict(ordered_list[:])


# #########################################################
# dictionary match with common l33t substitutions
# #########################################################


def relevant_l33t_subtable(password, table):
    password_chars = {}
    for c in password:
        password_chars[c] = True
    subtable = {}
    for letter, subs in table.items():
        relevant_subs = [sub for sub in subs if sub in password_chars]
        if len(relevant_subs) > 0:
            subtable[letter] = relevant_subs
    return subtable


def enumerate_l33t_subs(table):
    """

    :param dict table:
    :return:
    """
    keys = [key for key in table.keys()]
    subs = [[[]]]

    def dedup(sub_list):
        """Delete duplicated data
        :param list of list sub_list:
        :return:
        """
        deduped = []
        members = {}
        for sub in sub_list:

            assoc = [[k, v] for k, v in sub]
            assoc.sort()
            label = "-".join([str(k) + "," + str(v) for k, v in assoc])
            if label not in members:
                members[label] = True
                deduped.append(sub)
        return deduped

    def helper(helper_keys):
        if len(helper_keys) == 0:
            return
        first_key = helper_keys[0]
        rest_keys = helper_keys[1:]
        next_subs = []
        for l33t_chr in table[first_key]:
            for sub in subs[0]:
                dup_l33t_index = [-1]
                for i in range(len(sub)):
                    if sub[i][0] == l33t_chr:
                        dup_l33t_index[0] = i
                        break
                if dup_l33t_index[0] == -1:
                    sub_extension = sub + [[l33t_chr, first_key]]
                    next_subs.append(sub_extension)
                else:
                    sub_alternative = sub[:]
                    sub_alternative.pop(dup_l33t_index[0])
                    sub_alternative.append([l33t_chr, first_key])
                    next_subs.append(sub)
                    next_subs.append(sub_alternative)
        subs[0] = dedup(next_subs)
        helper(rest_keys)

    helper(keys)
    sub_dicts = []  # convert from assoc lists to dicts
    for sub in subs[0]:
        sub_dict = {}
        for l33t_c, c in sub:
            sub_dict[l33t_c] = c
        sub_dicts.append(sub_dict)
    return sub_dicts


def l33t_match(password, _ranked_dictionaries=RANKED_DICTIONARIES, _l33t_table=L33T_TABLE):
    matches = []
    for sub in enumerate_l33t_subs(relevant_l33t_subtable(password, _l33t_table)):
        if is_empty(sub):
            break
        subbed_password = translate(password, sub)
        for match in dictionary_match(subbed_password, _ranked_dictionaries):
            token = password[match["i"]:match["j"]+1]
            if token.lower() == match["matched_word"]:
                continue
            match_sub = {}
            for subbed_c, c in sub.items():
                if token.find(subbed_c) == -1:
                    continue
                match_sub[subbed_c] = c

            match["l33t"] = True
            match["token"] = token
            match["sub"] = match_sub
            match["sub_display"] = ", ".join(["{} -> {}".format(k, v) for k, v in match_sub.items()])
            matches.append(match)
    return sorted([m for m in matches if len(m["token"]) > 1], key=lambda x: (x['i'], x['j']))


# #########################################################
# spatial match (qwerty/dvorak/keypad)
# #########################################################

SHIFTED_RX = '[~!@#$%^&*()_+QWERTYUIOP{}|ASDFGHJKL:"ZXCVBNM<>?]'


def spatial_match(password, _graphs=GRAPHS):
    matches = []
    for graph_name, graph in _graphs.items():
        matches.extend(spatial_match_helper(password, graph, graph_name))
    return sorted(matches, key=lambda x: (x["i"], x["j"]))


def spatial_match_helper(password, graph, graph_name):
    matches = []
    i = 0
    while i < len(password) - 1:
        j = i + 1
        last_direction = None
        turns = 0

        if graph_name in ['qwerty', 'dvorak'] and re.findall(SHIFTED_RX, password[i]):
            # initial character is shifted
            shifted_count = 1
        else:
            shifted_count = 0

        while True:
            prev_char = password[j - 1]
            found = False
            found_direction = -1
            cur_direction = -1
            adjacents = graph[prev_char] if prev_char in graph else []
            # consider growing pattern by one character if j hasn't gone over the edge
            if j < len(password):
                cur_char = password[j]
                for adj in adjacents:
                    cur_direction += 1
                    if adj and adj.find(cur_char) != -1:
                        found = True
                        found_direction = cur_direction
                        if adj.find(cur_char) == 1:
                            shifted_count += 1
                        if last_direction != found_direction:
                            turns += 1
                            last_direction = found_direction
                        break

            if found:
                j += 1

            else:
                if (j - i) > 2:
                    matches.append({
                        "pattern": "spatial",
                        "i": i,
                        "j": j - 1,
                        "token": password[i:j],
                        "graph": graph_name,
                        "turns": turns,
                        "shifted_count": shifted_count
                    })
                i = j
                break

    return matches


# #########################################################
# repeats (aaa, abcabcabc) and sequences (abcdef)
# #########################################################

def repeat_match(password):
    matches = []
    greedy = r"(.+)\1+"
    lazy = r"(.+?)\1+"
    lazy_anchored = r"^(.+?)\1+$"
    lastIndex = 0

    while lastIndex < len(password):
        greedy_match = re.search(greedy, password[lastIndex:])
        lazy_match = re.search(lazy, password[lastIndex:])

        if greedy_match is None:
            break

        if len(greedy_match.group(0)) > len(lazy_match.group(0)):
            match = greedy_match
            base_token = re.search(lazy_anchored, match.group(0)).group(1)

        else:
            match = lazy_match
            base_token = match.group(1)

        i, j = [match.start() + lastIndex, match.start() + len(match.group(0)) - 1 + lastIndex]

        # TODO: Implement base analysis
        base_analysis = scoring.most_guessable_match_sequence(base_token, omnimatch(base_token))
        base_matches = base_analysis["match_sequence"] if "match_sequence" in base_analysis and base_analysis["match_sequence"] is not None else None
        base_guesses = base_analysis["guesses"]
        matches.append({
            "pattern": "repeat",
            "i": i,
            "j": j,
            "token": match.group(0),
            "base_token": base_token,
            "base_guesses": base_guesses,
            "base_matches": base_matches,
            "repeat_count": len(match.group(0)) / len(base_token)
        })
        lastIndex = j + 1
    return matches


def sequence_match(password):
    matches = []
    for sequence_name, sequence in SEQUENCES.items():
        for direction in [1, -1]:
            i = 0
            while i < len(password):
                if password[i] not in sequence:
                    i += 1
                    continue
                j = i + 1
                sequence_position = sequence.find(password[i])

                while j < len(password):
                    next_sequence_position = mod(sequence_position + direction, len(sequence))
                    if sequence.find(password[j]) != next_sequence_position:
                        break
                    j += 1
                    sequence_position = next_sequence_position
                j -= 1
                if (j - i + 1) > 1:
                    matches.append({
                        "pattern": "sequence",
                        "i": i,
                        "j": j,
                        "token": password[i:j+1],
                        "sequence_name": sequence_name,
                        "sequence_space": len(sequence),
                        "ascending": direction == 1
                    })
                i = j + 1

    return sorted(matches, key=lambda x: (x['i'], x['j']))


# #########################################################
# regex matching
# #########################################################

def regex_match(password, _regexen=REGEXEN):
    matches = []
    for name, regex in _regexen.items():
        rx_matches = re.finditer(regex, password)
        for rx_match in rx_matches:
            token = rx_match.group(0)
            matches.append({
                "pattern": "regex",
                "token": token,
                "i": rx_match.start(),
                "j": rx_match.start() + len(rx_match.group(0)) - 1,
                "regex_name": name,
                "regex_match": rx_match
            })

    precedence_map = {}

    def get_key(m):
        return "{}-{}".format(m["i"], m["j"])

    for match in matches:
        key = get_key(match)
        precedence = REGEX_PRECEDENCE[match["regex_name"]]
        if key in precedence_map:
            highest_precedence = precedence_map[key]
            if highest_precedence >= precedence:
                continue
        precedence_map[key] = precedence

    return sorted(
        [m for m in matches if precedence_map[get_key(m)] == REGEX_PRECEDENCE[m["regex_name"]]],
        key=lambda x: (x["i"], x["j"])
    )


# #########################################################
# date matching
# #########################################################

def date_match(password):
    matches = []
    maybe_date_no_separator = r"^\d{4,8}$"
    maybe_date_with_separator = r"^(\d{1,4})([\s/\\_.-])(\d{1,2})\2(\d{1,4})$"

    # dates without separators are between length 4 '1985' and 8 '29051985'
    for i in range(len(password) - 3):
        for j in range(i+3, i+8):
            if j >= len(password):
                break
            token = password[i:j+1]
            rx_match = re.match(maybe_date_no_separator, token)
            if rx_match is None:
                continue
            candidates = []
            for k, l in DATE_SPLITS[len(token)]:
                dmy = map_ints_to_dmy([
                    int(token[0:k]),
                    int(token[k:l]),
                    int(token[l:])
                ])
                if dmy is not None:
                    candidates.append(dmy)

            if not len(candidates) > 0:
                continue

            def metric(candidate_date):
                return math.fabs(candidate_date['year'] - scoring.REFERENCE_YEAR)

            best_candidate = candidates[0]
            min_distance = metric(candidates[0])

            for candidate in candidates[1:]:
                distance = metric(candidate)
                if distance < min_distance:
                    best_candidate = candidate
                    min_distance = distance

            matches.append({
                "pattern": "date",
                "token": token,
                "i": i,
                "j": j,
                "separator": "",
                "year": best_candidate["year"],
                "month": best_candidate["month"],
                "day": best_candidate["day"]
            })

    # dates with separators are between length 6 '5/9/91' and 10 '05/29/1985'
    for i in range(len(password) - 5):
        for j in range(i+5, i+10):
            if j >= len(password):
                break
            token = password[i:j+1]
            rx_match = re.match(maybe_date_with_separator, token)
            if rx_match is None:
                continue

            dmy = map_ints_to_dmy([
                int(rx_match.group(1)),
                int(rx_match.group(3)),
                int(rx_match.group(4))
            ])

            if dmy is None:
                continue

            matches.append({
                "pattern": "date",
                "token": token,
                "i": i,
                "j": j,
                "separator": rx_match.group(2),
                "year": dmy["year"],
                "month": dmy["month"],
                "day": dmy["day"],
            })

    def del_submatch(match):
        is_submatch = False
        for other_match in matches:
            if match is other_match:
                continue
            if other_match["i"] <= match["i"] and other_match["j"] >= match["j"]:
                is_submatch = True
                break
        return not is_submatch

    return sorted(filter(del_submatch, matches), key=lambda d: tuple(d[k] for k in sorted(d)))


def map_ints_to_dmy(int_list):
    """Convert list of integer into dict of (year, month, day)
    :param list of int int_list:
    :retype: dict
    """
    if int_list[1] > 31 or int_list[1] <= 0:
        return None

    over_12 = 0
    over_31 = 0
    under_1 = 0
    for i in int_list:
        if 99 < i < DATE_MIN_YEAR or i > DATE_MAX_YEAR:
            return None

        if i > 31:
            over_31 += 1
        if i > 12:
            over_12 += 1
        if i <= 0:
            under_1 += 1

    if over_31 >= 2 or over_12 == 3 or under_1 >= 2:
        return None

    possible_year_splits = [
        [int_list[2], int_list[:2]],  # year last
        [int_list[0], int_list[1:]]   # year first
    ]
    for [y, rest] in possible_year_splits:
        if DATE_MIN_YEAR <= y <= DATE_MAX_YEAR:
            dm = map_ints_to_dm(rest)
            if dm is not None:
                return {
                    "year": y,
                    "month": dm["month"],
                    "day": dm["day"]
                }
            else:
                return None

    for [y, rest] in possible_year_splits:
        dm = map_ints_to_dm(rest)
        if dm is not None:
            y = two_to_four_digit_year(y)
            return {
                "year": y,
                "month": dm["month"],
                "day": dm["day"]
            }


def map_ints_to_dm(int_list):
    """Convert list of integer into dict of (day, month)
    :param list int_list:
    :rtype: dict
    """
    for d, m in [int_list, reversed(int_list)]:
        if 1 <= d <= 31 and 1 <= m <= 12:
            return {
                "day": d,
                "month": m
            }
    return None


def two_to_four_digit_year(year):
    """Convert two-digits year into four-digits year
    :param int year:
    :rtype: int
    """
    if year > 99:
        return year
    elif year > 50:
        # 85 -> 1985
        return year + scoring.REFERENCE_YEAR - 100
    else:
        # 15 -> 2015
        return year + scoring.REFERENCE_YEAR




def estimate_attack_times(guesses):
    crack_times_seconds = {
        "online_throttling_100_per_hour": guesses / (100.0 / 3600.0),
        "online_no_throttling_10_per_second": guesses / 1e2,
        "offline_slow_hashing_1e4_per_second": guesses / 1e4,
        "offline_fast_hashing_1e10_per_second": guesses / 1e10
    }

    crack_times_display = {}
    for scenario, seconds in crack_times_seconds.items():
        crack_times_display[scenario] = display_time(seconds)

    return {
        "crack_times_seconds": crack_times_seconds,
        "crack_times_display": crack_times_display,
        "score": guesses_to_score(guesses)
    }


def guesses_to_score(guesses):
    DELTA = 5
    if guesses < (1e3 + DELTA):
        # risky password: "too guessable"
        return 0
    elif guesses < (1e6 + DELTA):
        # modest protection from throttled online attacks: "very guessable"
        return 1
    elif guesses < (1e8 + DELTA):
        # modest protection from unthrottled online attacks: "somewhat guessable"
        return 2
    elif guesses < (1e10 + DELTA):
        # modest protection from offline attacks: "safely unguessable"
        # assuming a salted, slow hash function like bcrypt, scrypt, PBKDF2, argon, etc
        return 3
    else:
        # strong protection from offline attacks under same scenario: "very unguessable"
        return 4


def display_time(seconds):
    minute = 60
    hour = minute * 60
    day = hour * 24
    month = day * 31
    year = month * 12
    century = year * 100

    if seconds < 1:
        display_num, display_str = (None, "less than a second")
    elif seconds < minute:
        base = round(seconds)
        display_num, display_str = (base, "{} second".format(base))
    elif seconds < hour:
        base = round(seconds / minute)
        display_num, display_str = (base, "{} minute".format(base))
    elif seconds < day:
        base = round(seconds / hour)
        display_num, display_str = (base, "{} hour".format(base))
    elif seconds < month:
        base = round(seconds / day)
        display_num, display_str = (base, "{} day".format(base))
    elif seconds < year:
        base = round(seconds / month)
        display_num, display_str = (base, "{} month".format(base))
    elif seconds < century:
        base = round(seconds / year)
        display_num, display_str = (base, "{} year".format(base))

    else:
        display_num, display_str = (None, "centuries")

    if display_num is not None and display_num != 1:
        display_str += "s"
    return display_str



from __future__ import absolute_import
from __future__ import print_function
import math
import re
from decimal import Decimal

from .adjacency_graphs import adjacency_graphs
from six.moves import range

# on qwerty, 'g' has degree 6, being adjacent to 'ftyhbv'. '\' has degree 1.
# this calculates the average over all keys.


def calc_average_degree(graph):
    average = 0
    for key, neighbors in graph.items():
        average += len([n for n in neighbors if n])
    average /= len([k for k in graph.keys()])
    return average

BRUTEFORCE_CARDINALITY = 10
MIN_GUESSES_BEFORE_GROWING_SEQUENCE = 10000
MIN_SUBMATCH_GUESSES_SINGLE_CHAR = 10
MIN_SUBMATCH_GUESSES_MULTI_CHAR = 50
REFERENCE_YEAR = 2000


def binom(n, k):
    """
    Returns binomial coefficient (n choose k).
    """
    # http://blog.plover.com/math/choose.html
    if k > n:
        return 0
    if k == 0:
        return 1
    result = 1
    for denom in range(1, k + 1):
        result *= n
        result /= denom
        n -= 1
    return result


def log10(n):
    """
    Returns logarithm of n in base 10.
    """
    return math.log(float(n), 10)


def log2(n):
    """
    Returns logarithm of n in base 2.
    """
    return math.log(n, 2)


def factorial(n):
    """
    Return factorial of n
    """
    if n < 2:
        return 1
    f = 1
    for i in range(1, n):
        f *= i
    return f


def insert_val_to_arr(array, index, value, default=None):
        if (len(array) - 1) > index:
            array[index] = value
        else:
            for i in range(len(array), index+1):
                array.append(default)
            array[index] = value


def most_guessable_match_sequence(password, matches, _exclude_additive=False):
    optimal_product = [[] for _ in range(len(password)+1)]
    backpointers = [[] for _ in range(len(password)+1)]

    max_l = 0
    optimal_l = None

    def make_bruteforce_match(i, j):
        return {
            "pattern": "bruteforce",
            "token": password[i:j+1],
            "i": i,
            "j": j
        }

    def score(guess_product, sequence_length):
        result = math.factorial(sequence_length) * guess_product
        if not _exclude_additive:
            result += math.pow(MIN_GUESSES_BEFORE_GROWING_SEQUENCE, sequence_length - 1)
        return result

    for k in range(len(password)):
        backpointers[k] = []
        optimal_product[k] = []
        optimal_score = Decimal("Infinity")

        for prev_l in range(max_l + 1):
            # for each new k, starting scenario to try to beat: bruteforce matches
            # involving the lowest-possible l. three cases:
            #
            # 1. all-bruteforce match (for length-1 sequences.)
            # 2. extending a previous bruteforce match
            #    (possible when optimal[k-1][l] ends in bf.)
            # 3. starting a new single-char bruteforce match
            #    (possible when optimal[k-1][l] exists but does not end in bf.)
            #
            # otherwise: there is no bruteforce starting scenario that might be better
            # than already-discovered lower-l sequences.
            consider_bruteforce = True
            bf_j = k
            try:
                if prev_l == 0:
                    bf_i = 0
                    new_l = 1
                elif "pattern" in backpointers[k-1][prev_l] and \
                        backpointers[k-1][prev_l]["pattern"] == "bruteforce":
                    bf_i = backpointers[k-1][prev_l]["i"]
                    new_l = prev_l
                elif backpointers[k-1][prev_l] is not None:
                    bf_i = k
                    new_l = prev_l + 1
                else:
                    # bf_i = 0
                    # new_l = None
                    consider_bruteforce = False
            except:
                # bf_i = 0
                # new_l = None
                consider_bruteforce = False

            if consider_bruteforce:
                bf_match = make_bruteforce_match(bf_i, bf_j)
                prev_j = k - len(bf_match["token"])  # end of preceeding match
                candidate_product = estimate_guesses(bf_match, password)
                if new_l > 1:
                    candidate_product *= optimal_product[prev_j][new_l - 1]
                candidate_score = score(candidate_product, new_l)

                if candidate_score < optimal_score:
                    optimal_score = candidate_score
                    # optimal_product[k][new_l] = candidate_product
                    insert_val_to_arr(optimal_product[k], new_l, candidate_product)
                    optimal_l = new_l
                    max_l = max(max_l, new_l)
                    # backpointers[k][new_l] = bf_match
                    insert_val_to_arr(backpointers[k], new_l, bf_match)

            # now try beating those bruteforce starting scenarios.
            # for each match m ending at k, see if forming a (prev_l + 1) sequence
            # ending at m is better than the current optimum.
            for match in matches:
                if match["j"] != k:
                    continue
                i, j = (match["i"], match["j"])

                if prev_l == 0:
                    # if forming a len-1 sequence [match], match.i must fully cover [0..k]
                    if i != 0:
                        continue
                else:
                    # it's only possible to form a new potentially-optimal sequence ending at
                    # match when there's an optimal length-prev_l sequence ending at match.i-1.
                    try:
                        if not optimal_product[i-1][prev_l]:
                            continue
                    except:
                        continue

                candidate_product = estimate_guesses(match, password)
                if prev_l > 0:
                    candidate_product *= optimal_product[i-1][prev_l]
                candidate_score = score(candidate_product, prev_l + 1)
                if candidate_score < optimal_score:
                    optimal_score = candidate_score
                    insert_val_to_arr(optimal_product[k], prev_l + 1, candidate_product)
                    # optimal_product[k][prev_l+1] = candidate_product
                    optimal_l = prev_l + 1
                    max_l = max(max_l, prev_l+1)
                    insert_val_to_arr(backpointers[k], prev_l + 1, match)
                    # backpointers[k][prev_l+1] = match

    # walk backwards and decode the optimal sequence
    match_sequence = []
    l = optimal_l
    k = len(password) - 1

    while k >= 0:
        match = backpointers[k][l]
        match_sequence.append(match)
        k = match["i"] - 1
        l -= 1
    match_sequence.reverse()

    # final result object
    return {
        "password": password,
        "guesses": optimal_score,
        "guesses_log10": log10(optimal_score),
        "sequence": match_sequence
    }


# ------------------------------------------------------------------------------
# guess estimation -- one function per match pattern ---------------------------
# ------------------------------------------------------------------------------
def estimate_guesses(match, password):
    if "guesses" in match and match["guesses"]:
        return match["guesses"]  # a match's guess estimate doesn't change. cache it.
    min_guesses = 1
    if len(match["token"]) < len(password):
        if len(match["token"]) == 1:
            min_guesses = MIN_SUBMATCH_GUESSES_SINGLE_CHAR
        else:
            min_guesses = MIN_SUBMATCH_GUESSES_MULTI_CHAR

    estimation_functions = {
        "bruteforce": bruteforce_guesses,
        "dictionary": dictionary_guesses,
        "spatial":    spatial_guesses,
        "repeat":     repeat_guesses,
        "sequence":   sequence_guesses,
        "regex":      regex_guesses,
        "date":       date_guesses
    }
    guesses = estimation_functions[match["pattern"]](match)
    if not isinstance(guesses, (int, float)):
        print("hoge")
    match["guesses"] = max(guesses, min_guesses)
    match["guesses_log10"] = log10(match["guesses"])
    return match["guesses"]


def bruteforce_guesses(match):
    guesses = math.pow(BRUTEFORCE_CARDINALITY, len(match["token"]))
    # small detail: make bruteforce matches at minimum one guess bigger than smallest allowed
    # submatch guesses, such that non-bruteforce submatches over the same [i..j] take precidence.
    if len(match["token"]) == 1:
        min_guesses = MIN_SUBMATCH_GUESSES_SINGLE_CHAR + 1
    else:
        min_guesses = MIN_SUBMATCH_GUESSES_MULTI_CHAR + 1
    return max(guesses, min_guesses)


def repeat_guesses(match):
    return match["base_guesses"] * match["repeat_count"]


def sequence_guesses(match):
    first_chr = match["token"][0]
    # lower guesses for obvious starting points
    if first_chr in ["a", "A", "z", "Z", "0", "1", "9"]:
        base_guesses = 4
    else:
        if first_chr.isdigit():
            base_guesses = 10  # digits
        else:
            # could give a higher base for uppercase,
            # assigning 26 to both upper and lower sequences is more conservative.
            base_guesses = 26
    if not match["ascending"]:
        # need to try a descending sequence in addition to every ascending sequence ->
        # 2x guesses
        base_guesses *= 2
    return base_guesses * len(match["token"])


MIN_YEAR_SPACE = 20


def regex_guesses(match):
    char_class_bases = {
        "alpha_lower":  26,
        "alpha_upper":  26,
        "alpha":        52,
        "alphanumeric": 62,
        "digits":       10,
        "symbols":      33
    }

    if "regex_name" in match and match["regex_name"] in char_class_bases:
        return math.pow(char_class_bases[match["regex_name"]], len(match["token"]))
    elif "regex_name" in match and match["regex_name"] == "recent_year":
        # conservative estimate of year space: num years from REFERENCE_YEAR.
        # if year is close to REFERENCE_YEAR, estimate a year space of MIN_YEAR_SPACE.
        year_space = math.fabs(int(match["regex_match"].group(0))) - REFERENCE_YEAR
        year_space = max(year_space, MIN_YEAR_SPACE)
        return year_space


def date_guesses(match):
    # base guesses: (year distance from REFERENCE_YEAR) * num_days * num_years
    year_space = max(math.fabs(match["year"] - REFERENCE_YEAR), MIN_YEAR_SPACE)
    guesses = year_space * 31 * 12

    # double for four-digit years
    if "has_full_year" in match and match["has_full_year"]:
        guesses *= 2
    # add factor of 4 for separator selection (one of ~4 choices)
    if "separator" in match and match["separator"]:
        guesses *= 4
    return guesses


KEYBOARD_AVERAGE_DEGREE = calc_average_degree(adjacency_graphs["qwerty"])
# slightly different for keypad/mac keypad, but close enough
KEYPAD_AVERAGE_DEGREE = calc_average_degree(adjacency_graphs["keypad"])

KEYBOARD_STARTING_POSITIONS = len([k for k, v in adjacency_graphs["qwerty"].items()])
KEYPAD_STARTING_POSITIONS = len([k for k, v in adjacency_graphs["keypad"].items()])


def spatial_guesses(match):
    if "graph" in match and match["graph"] in ['qwerty', 'dvorak']:
        s = KEYBOARD_STARTING_POSITIONS
        d = KEYBOARD_AVERAGE_DEGREE
    else:
        s = KEYPAD_STARTING_POSITIONS
        d = KEYPAD_AVERAGE_DEGREE
    guesses = 0
    L = len(match["token"])
    t = match["turns"]
    # estimate the number of possible patterns w/ length L or less with t turns or less.
    for i in range(2, L + 1):
        possible_turns = min(t, i - 1)
        for j in range(1, possible_turns + 1):
            guesses += binom(i - 1, j - 1) * s * math.pow(d, j)

    # add extra guesses for shifted keys. (% instead of 5, A instead of a.)
    # math is similar to extra guesses of l33t substitutions in dictionary matches.
    if "shifted_count" in match and match["shifted_count"]:
        S = match["shifted_count"]
        U = len(match["token"]) - match["shifted_count"]  # unshifted count
        if S == 0 or U == 0:
            guesses *= 2
        else:
            shifted_variations = 0
            for i in range(1, min(S, U) + 1):
                shifted_variations += binom(S + U, i)
            guesses *= shifted_variations
    return guesses


def dictionary_guesses(match):
    match["base_guesses"] = match["rank"]  # keep these as properties for display purposes
    match["uppercase_variations"] = uppercase_variations(match)
    match["l33t_variations"] = l33t_variations(match)
    reversed_variations = 2 if "reversed" in match and match["reversed"] else 1
    return match["base_guesses"] * match["uppercase_variations"] * match["l33t_variations"] * reversed_variations


START_UPPER = r"^[A-Z][^A-Z]+$"
END_UPPER = r"^[^A-Z]+[A-Z]$"
ALL_UPPER = r"^[^a-z]+$"
ALL_LOWER = r"^[^A-Z]+$"
NO_LETTER = r"^$"


def uppercase_variations(match):
    word = match["token"]
    if re.search(ALL_LOWER, word) or re.search(NO_LETTER, word):
        return 1

    # a capitalized word is the most common capitalization scheme,
    # so it only doubles the search space (uncapitalized + capitalized).
    # allcaps and end-capitalized are common enough too, underestimate as 2x factor to be safe.
    for regex in [START_UPPER, END_UPPER, ALL_UPPER]:
        if re.search(regex, word):
            return 2

    # otherwise calculate the number of ways to capitalize U+L uppercase+lowercase letters
    # with U uppercase letters or less. or, if there's more uppercase than lower (for eg. PASSwORD),
    # the number of ways to lowercase U+L letters with L lowercase letters or less.
    U = len([c for c in word if re.match(u"[A-Z]", c)])
    L = len([c for c in word if re.match(u"[a-z]", c)])

    variations = 0
    for i in range(1, min(U, L) + 1):
        variations += binom(U + L, i)
    return variations


def l33t_variations(match):
    if "l33t" not in match or not match["l33t"]:
        return 1
    variations = 1

    for subbed, unsubbed in match["sub"].items():
        # lower-case match.token before calculating: capitalization shouldn't affect l33t calc.
        c_list = match["token"].lower()
        num_subbed = len([c for c in c_list if c == subbed])
        num_unsubbed = len([c for c in c_list if c == unsubbed])

        if num_subbed == 0 or num_unsubbed == 0:
            # for this sub, password is either fully subbed (444) or fully unsubbed (aaa)
            # treat that as doubling the space (attacker needs to try fully subbed chars in addition to
            # unsubbed.)
            variations *= 2

        else:
            # this case is similar to capitalization:
            # with aa44a, U = 3, S = 2, attacker needs to try unsubbed + one sub + two subs
            p = min(num_unsubbed, num_subbed)
            possibilities = 0
            for i in range(1, p+1):
                possibilities += binom(num_unsubbed + num_subbed, i)
            variations *= possibilities

    return variations



# -*- coding: utf-8 -*-
from .main import zxcvbn

__title__ = "pyzxcvbn"
__version__ = "0.6.0"
__author__ = "Takuro Wada"
__author_email__ = "taxpon@gmail.com"
__url__ = "https://github.com/taxpon/pyzxcvbn"
__license__ = "MIT"



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import re

from . import scoring

default_feedback = {
    "warning": "",
    "suggestions": [
        "Use a few words, avoid common phrases",
        "No need for symbols, digits, or uppercase letters"
    ]
}


def get_feedback(score, sequence):
    # starting feedback
    if len(sequence) == 0:
        return default_feedback

    # no feedback if score is good or great.
    if score > 2:
        return {
            "warning": "",
            "suggestions": []
        }

    # tie feedback to the longest match for longer sequences
    longest_match = sequence[0]
    for match in sequence[1:]:
        if len(match["token"]) > len(longest_match["token"]):
            longest_match = match
    feedback = get_match_feedback(longest_match, len(sequence) == 1)
    extra_feedback = "Add another word or two. Uncommon words are better."

    if feedback:
        feedback["suggestions"].insert(1, extra_feedback)
        if "warning" not in feedback or feedback["warning"] is None:
            feedback["warning"] = ""
    else:
        feedback = {
            "warning": "",
            "suggestions": [extra_feedback]
        }

    return feedback


def get_match_feedback(match, is_sole_match):

    if "pattern" not in match:
        return

    elif match["pattern"] == "dictionary":
        return get_dictionary_match_feedback(match, is_sole_match)

    elif match["pattern"] == "spatial":
        if "turns" in match and match["turns"] == 1:
            warning = "Straight rows of keys are easy to guess"
        else:
            warning = "Short keyboard patterns are easy to guess"

        return {
            "warning": warning,
            "suggestions": [
                "Use a longer keyboard pattern with more turns"
            ]
        }

    elif match["pattern"] == "repeat":
        if len(match["base_token"]) == 1:
            warning = "Repeats like 'aaa' are easy to guess"
        else:
            warning = "Repeats like 'abcabcabc' are only slightly harder to guess than 'abc'"
        return {
            "warning": warning,
            "suggestions": [
                "Avoid repeated words and characters"
            ]
        }

    elif match["pattern"] == "sequence":
        return {
            "warning": "Sequences like abc or 6543 are easy to guess",
            "suggestions": [
                "Avoid sequences"
            ]
        }

    elif match["pattern"] == "regex":
        if match["regex_name"] == "recent_year":
            return {
                "warning": "Recent years are easy to guess",
                "suggestions": [
                    "Avoid recent years",
                    "Avoid years that are associated with you"
                ]
            }

    elif match["pattern"] == "date":
        return {
            "warning": "Dates are often easy to guess",
            "suggestions": [
                "Avoid dates and years that are associated with you"
            ]
        }


def get_dictionary_match_feedback(match, is_sole_match):

    warning = ""
    if "dictionary_name" not in match:
        warning = ""

    elif match["dictionary_name"] == "passwords":
        if is_sole_match and \
                (("l33t" not in match) or not match["l33t"]) and \
                (("reversed" not in match) or not match["reversed"]):
            if match["rank"] <= 10:
                warning = "This is a top-10 common password"
            elif match["rank"] <= 100:
                warning = "This is a top-100 common password"
            else:
                warning = "This is a very common password"
        elif match["guesses_log10"] <= 4:
            warning = "This is similar to a commonly used password"

    elif match["dictionary_name"] == "english":
        if is_sole_match:
            warning = "A word by itself is easy to guess"

    elif match["dictionary_name"] in ["surnames", "male_names", "female_names"]:
        if is_sole_match:
            warning = "Names and surnames by themselves are easy to guess"
        else:
            warning = "Common names and surnames are easy to guess"

    else:
        warning = ""

    suggestions = []
    word = match["token"]

    if re.match(scoring.START_UPPER, word):
        suggestions.append("Capitalization doesn't help very much")
    elif re.match(scoring.ALL_UPPER, word):
        suggestions.append("All-uppercase is almost as easy to guess as all-lowercase")

    if "reversed" in match and match["reversed"] and len(match["token"]) >= 4:
        suggestions.append("Reversed words aren't much harder to guess")

    if "l33t" in match and match["l33t"]:
        suggestions.append("Predictable substitutions like '@' instead of 'a' don't help very much")

    return {
        "warning": warning,
        "suggestions": suggestions
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import datetime

from . import matching
from . import scoring
from . import time_estimates
from . import feedback


def zxcvbn(password, user_inputs=()):
    """Measure strength of the password
    :param str password:
    :param list user_inputs:
    :rtype: dict
    """
    start = datetime.datetime.now()
    # reset the user inputs matcher on a per-request basis to keep things stateless
    sanitized_inputs = []
    for arg in user_inputs:
        if isinstance(arg, (str, int, bool)):
            sanitized_inputs.append(str(arg).lower())
    matching.set_user_input_dictionary(sanitized_inputs)
    matches = matching.omnimatch(password)
    result = scoring.most_guessable_match_sequence(password, matches)
    result["calc_time"] = datetime.datetime.now() - start
    attack_time = time_estimates.estimate_attack_times(result["guesses"])
    for prop, val in attack_time.items():
        result[prop] = val
    result["feedback"] = feedback.get_feedback(result["score"], result["sequence"])
    return result



from setuptools import setup, find_packages
import sys

import pyzxcvbn

sys.path.append("./pyzxcvbn")
sys.path.append('./tests.py')

setup(
    name=pyzxcvbn.__title__,
    packages=find_packages(),
    version=pyzxcvbn.__version__,
    author=pyzxcvbn.__author__,
    author_email=pyzxcvbn.__author_email__,
    short_description="Python version zxcvbn",
    long_description="""\
=================
pyzxcvbn
=================

Python version `zxcvbn <https://github.com/dropbox/zxcvbn>`_.
There are the same test scripts (but written in not coffee script but python) as the original repository in order to ensure this library will move in the same way.

Install
-------
::
    $ pip install pyzxcvbn

""",
    url=pyzxcvbn.__url__,
    license=pyzxcvbn.__license__,
    classmethod=[
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python",
    ],
    test_suite='tests.suite',
    install_requires=[
        'six',
    ],
)
